@@ -1,6 +1,5 @@
 from glob import glob
 import re, os, sys
-from cython import set
 
 
 from distutils.extension import Extension
@@ -8,7 +7,19 @@ from distutils.extension import Extension
 from Cython import Utils
 from Cython.Compiler.Main import Context, CompilationOptions, default_options
 
-# Unfortunately, Python 2.3 doesn't support decorators.
+def cached_function(f):
+    cache_name = '__%s_cache' % f.__name__
+    def wrapper(*args):
+        cache = getattr(f, cache_name, None)
+        if cache is None:
+            cache = {}
+            setattr(f, cache_name, cache)
+        if args in cache:
+            return cache[args]
+        res = cache[args] = f(*args)
+        return res
+    return wrapper
+
 def cached_method(f):
     cache_name = '__%s_cache' % f.__name__
     def wrapper(self, *args):
@@ -21,9 +32,37 @@ def cached_method(f):
         res = cache[args] = f(self, *args)
         return res
     return wrapper
-
+    
+def extended_iglob(pattern):
+    if '**/' in pattern:
+        seen = set()
+        first, rest = pattern.split('**/', 1)
+        if first == '':
+            first = '.'
+        for root in glob(first + "/"):
+            for path in extended_iglob(os.path.join(root, rest)):
+                if path not in seen:
+                    seen.add(path)
+                    yield path
+            for path in extended_iglob(os.path.join(root, '*', '**', rest)):
+                if path not in seen:
+                    seen.add(path)
+                    yield path
+    else:
+        for path in glob(pattern):
+            yield path
 
 def parse_list(s):
+    """
+    >>> parse_list("a b c")
+    ['a', 'b', 'c']
+    >>> parse_list("[a, b, c]")
+    ['a', 'b', 'c']
+    >>> parse_list('a " " b')
+    ['a', ' ', 'b']
+    >>> parse_list('[a, ",a", "a,", ",", ]')
+    ['a', ',a', 'a,', ',']
+    """
     if s[0] == '[' and s[-1] == ']':
         s = s[1:-1]
         delimiter = ','
@@ -32,12 +71,11 @@ def parse_list(s):
     s, literals = strip_string_literals(s)
     def unquote(literal):
         literal = literal.strip()
-        if literal[0] == "'":
+        if literal[0] in "'\"":
             return literals[literal[1:-1]]
         else:
             return literal
-
-    return [unquote(item) for item in s.split(delimiter)]
+    return [unquote(item) for item in s.split(delimiter) if item.strip()]
 
 transitive_str = object()
 transitive_list = object()
@@ -231,7 +269,10 @@ def parse_dependencies(source_filename):
     if '\t' in source:
         source = source.replace('\t', ' ')
     # TODO: pure mode
-    dependancy = re.compile(r"(cimport +([0-9a-zA-Z_.]+)\b)|(from +([0-9a-zA-Z_.]+) +cimport)|(include +'([^']+)')|(cdef +extern +from +'([^']+)')")
+    dependancy = re.compile(r"(cimport +([0-9a-zA-Z_.]+)\b)|"
+                             "(from +([0-9a-zA-Z_.]+) +cimport)|"
+                             "(include +['\"]([^'\"]+)['\"])|"
+                             "(cdef +extern +from +['\"]([^'\"]+)['\"])")
     cimports = []
     includes = []
     externs  = []
@@ -254,12 +295,11 @@ class DependencyTree(object):
         self.context = context
         self._transitive_cache = {}
 
-    #@cached_method
+    @cached_method
     def parse_dependencies(self, source_filename):
         return parse_dependencies(source_filename)
-    parse_dependencies = cached_method(parse_dependencies)
-
-    #@cached_method
+    
+    @cached_method
     def cimports_and_externs(self, filename):
         cimports, includes, externs = self.parse_dependencies(filename)[:3]
         cimports = set(cimports)
@@ -275,25 +315,22 @@ class DependencyTree(object):
             else:
                 print("Unable to locate '%s' referenced from '%s'" % (filename, include))
         return tuple(cimports), tuple(externs)
-    cimports_and_externs = cached_method(cimports_and_externs)
 
     def cimports(self, filename):
         return self.cimports_and_externs(filename)[0]
 
-    #@cached_method
+    @cached_method
     def package(self, filename):
-        dir = os.path.dirname(filename)
-        if os.path.exists(os.path.join(dir, '__init__.py')):
+        dir = os.path.dirname(os.path.abspath(filename))
+        if dir != filename and os.path.exists(os.path.join(dir, '__init__.py')):
             return self.package(dir) + (os.path.basename(dir),)
         else:
             return ()
-    package = cached_method(package)
 
-    #@cached_method
+    @cached_method
     def fully_qualifeid_name(self, filename):
         module = os.path.splitext(os.path.basename(filename))[0]
         return '.'.join(self.package(filename) + (module,))
-    fully_qualifeid_name = cached_method(fully_qualifeid_name)
 
     def find_pxd(self, module, filename=None):
         if module[0] == '.':
@@ -306,20 +343,19 @@ class DependencyTree(object):
         return self.context.find_pxd_file(module, None)
     find_pxd = cached_method(find_pxd)
 
-    #@cached_method
+    @cached_method
     def cimported_files(self, filename):
         if filename[-4:] == '.pyx' and os.path.exists(filename[:-4] + '.pxd'):
             self_pxd = [filename[:-4] + '.pxd']
         else:
             self_pxd = []
-        a = self.cimports(filename)
+        a = list(x for x in self.cimports(filename) if x.split('.')[0] != 'cython')
         b = filter(None, [self.find_pxd(m, filename) for m in self.cimports(filename)])
-        if len(a) - int('cython' in a) != len(b):
+        if len(a) != len(b):
             print("missing cimport", filename)
             print("\n\t".join(a))
             print("\n\t".join(b))
         return tuple(self_pxd + filter(None, [self.find_pxd(m, filename) for m in self.cimports(filename)]))
-    cimported_files = cached_method(cimported_files)
 
     def immediate_dependencies(self, filename):
         all = list(self.cimported_files(filename))
@@ -327,10 +363,9 @@ class DependencyTree(object):
             all.append(os.path.normpath(os.path.join(os.path.dirname(filename), extern)))
         return tuple(all)
 
-    #@cached_method
+    @cached_method
     def timestamp(self, filename):
         return os.path.getmtime(filename)
-    timestamp = cached_method(timestamp)
 
     def extract_timestamp(self, filename):
         # TODO: .h files from extern blocks
@@ -397,7 +432,7 @@ def create_extension_list(patterns, exclude=[], ctx=None, aliases=None):
     if not isinstance(exclude, list):
         exclude = [exclude]
     for pattern in exclude:
-        to_exclude.update(glob(pattern))
+        to_exclude.update(extended_iglob(pattern))
     if not isinstance(patterns, list):
         patterns = [patterns]
     module_list = []
@@ -420,7 +455,7 @@ def create_extension_list(patterns, exclude=[], ctx=None, aliases=None):
             exn_type = template.__class__
         else:
             raise TypeError(pattern)
-        for file in glob(filepattern):
+        for file in extended_iglob(filepattern):
             if file in to_exclude:
                 continue
             pkg = deps.package(file)
@@ -446,7 +481,7 @@ def create_extension_list(patterns, exclude=[], ctx=None, aliases=None):
     return module_list
 
 # This is the user-exposed entry point.
-def cythonize(module_list, exclude=[], nthreads=0, aliases=None, quiet=False, **options):
+def cythonize(module_list, exclude=[], nthreads=0, aliases=None, quiet=False, force=False, **options):
     if 'include_path' not in options:
         options['include_path'] = ['.']
     c_options = CompilationOptions(**options)
@@ -482,7 +517,7 @@ def cythonize(module_list, exclude=[], nthreads=0, aliases=None, quiet=False, **
                 else:
                     dep_timestamp, dep = deps.newest_dependency(source)
                     priority = 2 - (dep in deps.immediate_dependencies(source))
-                if c_timestamp < dep_timestamp:
+                if force or c_timestamp < dep_timestamp:
                     if not quiet:
                         if source == dep:
                             print("Compiling %s because it changed." % source)
@@ -505,14 +540,16 @@ def cythonize(module_list, exclude=[], nthreads=0, aliases=None, quiet=False, **
             nthreads = 0
     if not nthreads:
         for priority, pyx_file, c_file, options in to_compile:
-            cythonize_one(pyx_file, c_file, options)
+            cythonize_one(pyx_file, c_file, quiet, options)
     return module_list
 
 # TODO: Share context? Issue: pyx processing leaks into pxd module
-def cythonize_one(pyx_file, c_file, options=None):
+def cythonize_one(pyx_file, c_file, quiet, options=None):
     from Cython.Compiler.Main import compile, default_options
     from Cython.Compiler.Errors import CompileError, PyrexError
 
+    if not quiet:
+        print "Cythonizing %s" % pyx_file
     if options is None:
         options = CompilationOptions(default_options)
     options.output_file = c_file
