@@ -1,4 +1,4 @@
-from Errors import error, warning, message, warn_once, InternalError
+from Errors import error, message
 import ExprNodes
 import Nodes
 import Builtin
@@ -7,35 +7,69 @@ from Cython import Utils
 from PyrexTypes import py_object_type, unspecified_type
 from Visitor import CythonTransform
 
-try:
-    set
-except NameError:
-    # Python 2.3
-    from sets import Set as set
-
 
 class TypedExprNode(ExprNodes.ExprNode):
-    # Used for declaring assignments of a specified type whithout a known entry.
+    # Used for declaring assignments of a specified type without a known entry.
     def __init__(self, type):
         self.type = type
 
 object_expr = TypedExprNode(py_object_type)
 
 class MarkAssignments(CythonTransform):
-    
-    def mark_assignment(self, lhs, rhs):
+
+    # tells us whether we're in a normal loop
+    in_loop = False
+
+    parallel_errors = False
+
+    def __init__(self, context):
+        super(CythonTransform, self).__init__()
+        self.context = context
+
+        # Track the parallel block scopes (with parallel, for i in prange())
+        self.parallel_block_stack = []
+
+    def mark_assignment(self, lhs, rhs, inplace_op=None):
         if isinstance(lhs, (ExprNodes.NameNode, Nodes.PyArgDeclNode)):
             if lhs.entry is None:
                 # TODO: This shouldn't happen...
                 return
             lhs.entry.assignments.append(rhs)
+
+            if self.parallel_block_stack:
+                parallel_node = self.parallel_block_stack[-1]
+                previous_assignment = parallel_node.assignments.get(lhs.entry)
+
+                # If there was a previous assignment to the variable, keep the
+                # previous assignment position
+                if previous_assignment:
+                    pos, previous_inplace_op = previous_assignment
+
+                    if (inplace_op and previous_inplace_op and
+                            inplace_op != previous_inplace_op):
+                        # x += y; x *= y
+                        t = (inplace_op, previous_inplace_op)
+                        error(lhs.pos,
+                              "Reduction operator '%s' is inconsistent "
+                              "with previous reduction operator '%s'" % t)
+                else:
+                    pos = lhs.pos
+
+                parallel_node.assignments[lhs.entry] = (pos, inplace_op)
+                parallel_node.assigned_nodes.append(lhs)
+
         elif isinstance(lhs, ExprNodes.SequenceNode):
             for arg in lhs.args:
                 self.mark_assignment(arg, object_expr)
         else:
             # Could use this info to infer cdef class attributes...
             pass
-    
+
+    def visit_WithTargetAssignmentStatNode(self, node):
+        self.mark_assignment(node.lhs, node.rhs)
+        self.visitchildren(node)
+        return node
+
     def visit_SingleAssignmentNode(self, node):
         self.mark_assignment(node.lhs, node.rhs)
         self.visitchildren(node)
@@ -46,9 +80,9 @@ class MarkAssignments(CythonTransform):
             self.mark_assignment(lhs, node.rhs)
         self.visitchildren(node)
         return node
-    
+
     def visit_InPlaceAssignmentNode(self, node):
-        self.mark_assignment(node.lhs, node.create_binop_node())
+        self.mark_assignment(node.lhs, node.create_binop_node(), node.operator)
         self.visitchildren(node)
         return node
 
@@ -56,6 +90,11 @@ class MarkAssignments(CythonTransform):
         # TODO: Remove redundancy with range optimization...
         is_special = False
         sequence = node.iterator.sequence
+        if isinstance(sequence, ExprNodes.SimpleCallNode):
+            function = sequence.function
+            if sequence.self is None and function.is_name:
+                if function.name == 'reversed' and len(sequence.args) == 1:
+                    sequence = sequence.args[0]
         if isinstance(sequence, ExprNodes.SimpleCallNode):
             function = sequence.function
             if sequence.self is None and function.is_name:
@@ -65,11 +104,12 @@ class MarkAssignments(CythonTransform):
                         self.mark_assignment(node.target, arg)
                     if len(sequence.args) > 2:
                         self.mark_assignment(
-                            node.target, 
+                            node.target,
                             ExprNodes.binop_node(node.pos,
                                                  '+',
                                                  sequence.args[0],
                                                  sequence.args[2]))
+
         if not is_special:
             # A for-loop basically translates to subsequent calls to
             # __getitem__(), so using an IndexNode here allows us to
@@ -80,6 +120,7 @@ class MarkAssignments(CythonTransform):
                 node.pos,
                 base = sequence,
                 index = ExprNodes.IntNode(node.pos, value = '0')))
+
         self.visitchildren(node)
         return node
 
@@ -87,19 +128,23 @@ class MarkAssignments(CythonTransform):
         self.mark_assignment(node.target, node.bound1)
         if node.step is not None:
             self.mark_assignment(node.target,
-                    ExprNodes.binop_node(node.pos, 
-                                         '+', 
-                                         node.bound1, 
+                    ExprNodes.binop_node(node.pos,
+                                         '+',
+                                         node.bound1,
                                          node.step))
         self.visitchildren(node)
         return node
 
+    def visit_WhileStatNode(self, node):
+        self.visitchildren(node)
+        return node
+
     def visit_ExceptClauseNode(self, node):
         if node.target is not None:
             self.mark_assignment(node.target, object_expr)
         self.visitchildren(node)
         return node
-    
+
     def visit_FromCImportStatNode(self, node):
         pass # Can't be assigned to...
 
@@ -121,6 +166,67 @@ class MarkAssignments(CythonTransform):
         self.visitchildren(node)
         return node
 
+    def visit_DelStatNode(self, node):
+        for arg in node.args:
+            self.mark_assignment(arg, arg)
+        self.visitchildren(node)
+        return node
+
+    def visit_ParallelStatNode(self, node):
+        if self.parallel_block_stack:
+            node.parent = self.parallel_block_stack[-1]
+        else:
+            node.parent = None
+
+        nested = False
+        if node.is_prange:
+            if not node.parent:
+                node.is_parallel = True
+            else:
+                node.is_parallel = (node.parent.is_prange or not
+                                    node.parent.is_parallel)
+                nested = node.parent.is_prange
+        else:
+            node.is_parallel = True
+            # Note: nested with parallel() blocks are handled by
+            # ParallelRangeTransform!
+            # nested = node.parent
+            nested = node.parent and node.parent.is_prange
+
+        self.parallel_block_stack.append(node)
+
+        nested = nested or len(self.parallel_block_stack) > 2
+        if not self.parallel_errors and nested and not node.is_prange:
+            error(node.pos, "Only prange() may be nested")
+            self.parallel_errors = True
+
+        if node.is_prange:
+            child_attrs = node.child_attrs
+            node.child_attrs = ['body', 'target', 'args']
+            self.visitchildren(node)
+            node.child_attrs = child_attrs
+
+            self.parallel_block_stack.pop()
+            if node.else_clause:
+                node.else_clause = self.visit(node.else_clause)
+        else:
+            self.visitchildren(node)
+            self.parallel_block_stack.pop()
+
+        self.parallel_errors = False
+        return node
+
+    def visit_YieldExprNode(self, node):
+        if self.parallel_block_stack:
+            error(node.pos, "Yield not allowed in parallel sections")
+
+        return node
+
+    def visit_ReturnStatNode(self, node):
+        node.in_parallel = bool(self.parallel_block_stack)
+        return node
+
+
 class MarkOverflowingArithmetic(CythonTransform):
 
     # It may be possible to integrate this with the above for
@@ -131,7 +237,7 @@ class MarkOverflowingArithmetic(CythonTransform):
     def __call__(self, root):
         self.env_stack = []
         self.env = root.scope
-        return super(MarkOverflowingArithmetic, self).__call__(root)        
+        return super(MarkOverflowingArithmetic, self).__call__(root)
 
     def visit_safe_node(self, node):
         self.might_overflow, saved = False, self.might_overflow
@@ -148,7 +254,7 @@ class MarkOverflowingArithmetic(CythonTransform):
         self.visitchildren(node)
         self.might_overflow = saved
         return node
-    
+
     def visit_FuncDefNode(self, node):
         self.env_stack.append(self.env)
         self.env = node.local_scope
@@ -162,34 +268,34 @@ class MarkOverflowingArithmetic(CythonTransform):
             if entry:
                 entry.might_overflow = True
         return node
-    
+
     def visit_BinopNode(self, node):
         if node.operator in '&|^':
             return self.visit_neutral_node(node)
         else:
             return self.visit_dangerous_node(node)
-    
+
     visit_UnopNode = visit_neutral_node
-    
+
     visit_UnaryMinusNode = visit_dangerous_node
-    
+
     visit_InPlaceAssignmentNode = visit_dangerous_node
-    
+
     visit_Node = visit_safe_node
-    
+
     def visit_assignment(self, lhs, rhs):
-        if (isinstance(rhs, ExprNodes.IntNode) 
+        if (isinstance(rhs, ExprNodes.IntNode)
                 and isinstance(lhs, ExprNodes.NameNode)
                 and Utils.long_literal(rhs.value)):
             entry = lhs.entry or self.env.lookup(lhs.name)
             if entry:
                 entry.might_overflow = True
-    
+
     def visit_SingleAssignmentNode(self, node):
         self.visit_assignment(node.lhs, node.rhs)
         self.visitchildren(node)
         return node
-    
+
     def visit_CascadedAssignmentNode(self, node):
         for lhs in node.lhs_list:
             self.visit_assignment(lhs, node.rhs)
@@ -215,8 +321,9 @@ class SimpleAssignmentTypeInferer(object):
     # TODO: Implement a real type inference algorithm.
     # (Something more powerful than just extending this one...)
     def infer_types(self, scope):
-        enabled = not scope.is_closure_scope and scope.directives['infer_types']
+        enabled = scope.directives['infer_types']
         verbose = scope.directives['infer_types.verbose']
+
         if enabled == True:
             spanning_type = aggressive_spanning_type
         elif enabled is None: # safe mode
@@ -225,8 +332,6 @@ class SimpleAssignmentTypeInferer(object):
             for entry in scope.entries.values():
                 if entry.type is unspecified_type:
                     entry.type = py_object_type
-            if scope.is_closure_scope:
-                fix_closure_entries(scope)
             return
 
         dependancies_by_entry = {} # entry -> dependancies
@@ -234,6 +339,10 @@ class SimpleAssignmentTypeInferer(object):
         ready_to_infer = []
         for name, entry in scope.entries.items():
             if entry.type is unspecified_type:
+                if entry.in_closure or entry.from_closure:
+                    # cross-closure type inference is not currently supported
+                    entry.type = py_object_type
+                    continue
                 all = set()
                 for expr in entry.assignments:
                     all.update(expr.type_dependencies(scope))
@@ -246,6 +355,7 @@ class SimpleAssignmentTypeInferer(object):
                             entries_by_dependancy[dep].add(entry)
                 else:
                     ready_to_infer.append(entry)
+
         def resolve_dependancy(dep):
             if dep in entries_by_dependancy:
                 for entry in entries_by_dependancy[dep]:
@@ -254,12 +364,13 @@ class SimpleAssignmentTypeInferer(object):
                     if not entry_deps and entry != dep:
                         del dependancies_by_entry[entry]
                         ready_to_infer.append(entry)
+
         # Try to infer things in order...
         while True:
             while ready_to_infer:
                 entry = ready_to_infer.pop()
                 types = [expr.infer_type(scope) for expr in entry.assignments]
-                if types:
+                if types and Utils.all(types):
                     entry.type = spanning_type(types, entry.might_overflow)
                 else:
                     # FIXME: raise a warning?
@@ -282,25 +393,12 @@ class SimpleAssignmentTypeInferer(object):
                             break
             if not ready_to_infer:
                 break
-                    
+
         # We can't figure out the rest with this algorithm, let them be objects.
         for entry in dependancies_by_entry:
             entry.type = py_object_type
             if verbose:
                 message(entry.pos, "inferred '%s' to be of type '%s' (default)" % (entry.name, entry.type))
-        #if scope.is_closure_scope:
-        #    fix_closure_entries(scope)
-
-def fix_closure_entries(scope):
-    """Temporary work-around to fix field types in the closure class
-    that were unknown at the time of creation and only determined
-    during type inference.
-    """
-    closure_entries = scope.scope_class.type.scope.entries
-    for name, entry in scope.entries.iteritems():
-        if name in closure_entries:
-            closure_entry = closure_entries[name]
-            closure_entry.type = entry.type
 
 def find_spanning_type(type1, type2):
     if type1 is type2:
@@ -346,7 +444,7 @@ def safe_spanning_type(types, might_overflow):
         # operations without other int types, so this is safe, too
         return result_type
     elif result_type.is_ptr and not (result_type.is_int and result_type.rank == 0):
-        # Any pointer except (signed|unsigned|) char* can't implicitly 
+        # Any pointer except (signed|unsigned|) char* can't implicitly
         # become a PyObject.
         return result_type
     elif result_type.is_cpp_class:
@@ -357,7 +455,7 @@ def safe_spanning_type(types, might_overflow):
         # used, won't arise in pure Python, and there shouldn't be side
         # effects, so I'm declaring this safe.
         return result_type
-    # TODO: double complex should be OK as well, but we need 
+    # TODO: double complex should be OK as well, but we need
     # to make sure everything is supported.
     elif result_type.is_int and not might_overflow:
         return result_type
